/* RS232 Example program.*/
/* Heather Chia */
/*	Handle incoming RS232 stuff.
*/
#include <p_std.h>
#include <p_file.h>
#include <p_gen.h>
#include <p_serial.h>
#ifdef	EPOC
    #include <epoc.h>
#endif

#define COM1   ('1' -1)  /* COM1 in DOS = TTY:0 in PLIB  */
#define COM2   ('2' -1)  /* COM2 in DOS = TTY:1 in PLIB  */

#ifdef EPOC
    #define TTY_LETTER	'A'      /* TTY:A or B  */
#else
    #define TTY_LETTER	('1 '-1)  /* TTY:0 or 1  --> COM1 or COM2 in DOS   */
#endif


#define BAUD_RATE	P_BAUD_9600
#define DATA_BITS       P_DATA_8        /* or _7        */

#ifdef  NO_PARITY
    #define TYPE_OF_PARITY		P_PAR_NONE
    #define PARITY_BIT_PRESENT     0               /* indicates absence of a parity bit    */
#else   /* not NO PARITY        */
    #define TYPE_OF_PARITY		P_PAR_NONE	/* indicates presence of a parity bit    */
    #define PARITY_BIT_PRESENT     P_PARITY        /* or _ODD _EVEN etc.   */
#endif

/*************** Handshaking options ***************************/

/* Choose zero or more handshaking options      */
/* Example: XON/XOF, no RTS/CTS                 */
#define OBEY_XON_XOFF   1       /* 1 if responding to XON/XOFF        */
#define SEND_XON_XOFF   1       /* 1 if sending XON/XOFF      */
#define OBEY_RTS_CTS    0       /* 1 if responding to RTS/CTS        */


/* #define ONE ONLY of the following....        */
#define DSR_NOT_USED            /* define this if not using DSR */
#undef	DSR_FAIL_IF_DROPS	/* .. .. if should fail when DSR line drops       */
#undef  DSR_HANDSHAKE		/* .. .. if using DSR for handshaking   */

/* #define ONE ONLY of the following....        */
#define DCD_NOT_USED            /* as above, but for DCD    */
#undef	DCD_FAIL_IF_DROPS
#undef  DCD_HANDSHAKE


/* Specify the `terminating characters'.
   When performing a read operation, you can specify a set of (zero or more)
   characters so that the read operation will complete when any of the
   characters in the set is received.

   A terminating character can only be in the range 00 - 31 inclusive. 
   You can specify any combination of terminating characters, or none at all.  

   The terminating character-set is specified in the call to p_iow(P_FSET)
   shortly after opening the port.
*/
#define SET_NTH_BIT(bit_num)  (1uL<<(bit_num))

#ifdef READING_UNTIL_TERMINATORS

    #define TERMINATING_CHAR_1 13	/* CR terminates a frame	*/
    #define TERMINATING_CHAR_2 10	/* or a LF terminates a frame	*/
    #define TERMINATING_CHAR_3 26	/* or a ctrl Z terminates a frame	*/

/* Combine the terminating characters together into a bit-set   */
    #define TERMINATOR_SET   (\
		  SET_NTH_BIT(TERMINATING_CHAR_1)	\
                | SET_NTH_BIT(TERMINATING_CHAR_2)	\
                | SET_NTH_BIT(TERMINATING_CHAR_3))

#else /* if no terminating characters at all   */
    #define TERMINATOR_SET   0uL
#endif




/* ------- Now build up the handshaking characteristics ---------    */
/* thou mayst cheerfully skip this bit                                  */
#define HANDSHAKING_TYPE        0       /* Handshaking not defined yet  */ 
#ifdef OBEY_XON_XOFF
    #define HANDSHAKING_TYPE        (P_OBEY_XOFF | HANDSHAKING_TYPE)
#endif
#ifdef SEND_XON_XOFF
    #define HANDSHAKING_TYPE        (P_SEND_XOFF | HANDSHAKING_TYPE)
#endif
#ifndef OBEY_RTS_CTS
    #define HANDSHAKING_TYPE        (P_IGN_CTS | HANDSHAKING_TYPE)
#endif

#ifdef  DSR_NOT_USED
    #define DSR_MASK        0
#endif
#ifdef	DSR_FAIL_IF_DROPS
    #define DSR_MASK        ( P_OBEY_DSR | P_FAIL_DSR )
#endif
#ifdef  DSR_HANDSHAKE
    #define DSR_MASK        P_OBEY_DSR
#endif
#ifdef  DCD_NOT_USED
    #define DCD_MASK        0
#endif
#ifdef	DCD_FAIL_IF_DROPS
    #define DCD_MASK        ( P_OBEY_DCD | P_FAIL_DCD )
#endif
#ifdef  DCD_HANDSHAKE
    #define DCD_MASK        P_OBEY_DCD
#endif



/* This is the serial characteristics definition block  */
LOCAL_D const P_SRCHAR srchar =
{
	BAUD_RATE, BAUD_RATE,
	DATA_BITS | PARITY_BIT_PRESENT,
	TYPE_OF_PARITY,
	HANDSHAKING_TYPE,
	0x11, 0x13,		/* defines what the XON and XOFF characters	*/
	0,			/* Flags = check parity	*/
	TERMINATOR_SET 
};	/* Termination mask	*/


LOCAL_D void *		RS232_channel;
LOCAL_D WORD		read_io_stat;
LOCAL_D IN_BUFFER *	in_pbuffer;
LOCAL_D UBYTE		read_requested = FALSE;
LOCAL_D UINT		number_of_foreign_signals = 0;

LOCAL_D WORD		write_io_stat;
LOCAL_D UBYTE		write_requested = FALSE;

typedef struct
    {
    UWORD	length;
    UBYTE	content[BUFFER_SIZE];
    } IN_BUFFER;

LOCAL_D IN_BUFFER		in_buffer;


LOCAL_C request_read(IN_BUFFER * pb);


GLDEF_C UINT ReadIsFinished(void)
{
    return  read_requested   /* and there was a read operation	*/
	    && read_io_stat != E_FILE_PENDING;	/* and it has completed	*/
}


GLDEF_C UINT HandleIncomingRSEvent(void)
/*  Handle incoming RS232 event.    

    Returns TRUE if it handled the completion of the read operation.
*/  
{
    INT err;
    INPUT_BUFFER *	pb;	/* buffer last read into	*/

    if (!read_requested || read_io_stat == E_FILE_PENDING)
        return FALSE;	/* nothing to do        */

    read_requested = FALSE;     /* no longer any read in progress       */

    err = read_io_stat;
    if (err >0)
        err = 0;
    switch (err) 
    {
    case E_FILE_RECORD:
      /* If no terminator encountered before	*/
      /* maximum number of bytes were received, then thats ok	*/
	   /* drop thru...       */
    case 0:			/* or if all ok	*/
       break;
    case E_FILE_OVERRUN:
    case E_GEN_OVER:
    case E_FILE_PARITY:
    case E_FILE_FRAME:
    case E_FILE_LINE:
    default:			/* or some other error	*/ 
       p_leave(err);   /* raise an error condition     */
    }
    return TRUE;	/* we handled the event	*/
}


GLDEF_C void InitialiseRS232(void)
{			/*    01234	*/
    static TEXT tty_name[] = "TTY:x";

    tty_name[4] = TTY_LETTER;

    number_of_foreign_signals = 0;
    RS232_channel = NULL;
    read_requested = FALSE;
    write_requested = FALSE;

    f_open(&RS232_channel, tty_name, -1);
    f_leave(p_iow(RS232_channel, P_FSET, &srchar));
}

LOCAL_C void request_read(IN_BUFFER * pb, UINT nbytes)
/* Start a read-operation - read up to nbytes into buffer at pb 
*/
{
    if (read_requested)
       p_panic(A_READ_IS_ALREADY_IN_PROGRESS);

    read_requested=TRUE;
    
    pb = &in_buffer[0];
    pb->length = nbytes;
    p_ioc(RS232_channel,
    	  P_FREAD,
    	  &read_io_stat,
    	  &pb->content[0],
    	  &pb->length);
}

LOCAL_C void process_recvd_bytes(IN_BUFFER *	pb)
{
    UBYTE *	p;
    UINT	i;

    for (i = pb->length, p = &pb->content[0]; i; i--, p++)
	-- process the byte --;
}


GLDEF_C UINT WriteIsFinished(void)
{
    return  write_requested   /* and there was a write operation	*/
	    && write_io_stat != E_FILE_PENDING;	/* and it has completed	*/
}

GLDEF_C UINT HandleWriteFinishedEvent(void)
/*  Returns TRUE if it handled the completion of the write operation.
*/  
{
    if (!write_requested || write_io_stat == E_FILE_PENDING)
       return FALSE;	/* nothing to do        */

    write_requested = FALSE;
    f_leave(write_io_stat);     /* leave if the write operation failed    */  
    return TRUE;	/* we did process the event	*/
}

/**************************************
RS232_write() Starts transmitting a block of data.  
*/
GLDEF_C void RS232_write(UBYTE * data, UINT nbytes)
{
    static UWORD length_word;

    write_requested = TRUE;
    length_word = nbytes;
    p_ioc(RS232_channel,
	    P_FWRITE,
  	    &write_io_stat,
	    data,
	    &length_word);
}

static UBYTE data_to_be_written[999];

GLDEF_C void ReadAndWrite(void)
{
    INT err;

    err = p_enter1(DoReadAndWrite);
    if (err <0)
    {
        sorry folks it failed
        cancel_read();
        cancel_write();
        close(&RS232_channel);
    }
    /* Replace any foreign signals      */
    IoSignalNtimes(number_of_foreign_signals);
    number_of_foreign_signals = 0;
}

GLDEF_C int DoReadAndWrite(void)
{
    INT number_of_foreign_signals;

    InitialiseRS232();

    request_read(IN_BUFFER * pb, BUFFER_SIZE); /* or as many bytes as you like  */
    RS232_write(&data_to_be_written[0], 99);
    FOREVER
    {
       p_iowait();
       if (HandleIncomingRSEvent())
       {
          process_recvd_bytes(pb);
	       break;
       }
       else if (HandleWriteFinishedEvent())
       {
          write finished - do something
          break;
	    }
       else
	    ++number_of_foreign_signals;
    }
    f_leave(cancel_read());
    f_leave(cancel_write());
    f_leave(close(&RS232_channel));
}

GLDEF_C int cancel_read(void)
{
    return cancel(read_outstanding, RS232_channel, &read_io_stat);
}

GLDEF_C int cancel_write(void)
{
    return cancel(write_outstanding, RS232_channel, &write_io_stat);
}

LOCAL_C int cancel(UINT is_outstanding, void * channel, WORD * pstat)
{
    INT ret;

    ret = *pstat;
    if (is_outstanding && ret == E_FILE_PENDING)
    {
       p_iow2(channel, P_FCANCEL);
       p_waitstat(pstat);
       ret = *pstat;
       if (ret == E_FILE_CANCEL)
	        ret = 0;
    }
    return ret;
}


GLDEF_C int close(void ** pchannel)
{
    INT ret;

    ret = p_close(*pchannel);
    *pchannel = NULL;
    return ret;
}
